import pandas as pd
import patools.trial as tr
import os
from collections import OrderedDict

class Dataset:

    def __init__(self, directory=os.getcwd(), nbs=False):
        self.trials = self.loadDataset(directory, nbs=nbs)
        self.df = pd.DataFrame(index=self.trials.keys())

    def loadDataset(self, directory, bidispersity=True, nbs=False):
        """
        Loads a full dataset of packings. The general directory struct that this
        will import is as follows:

        Particle Number -> Radius Ratio -> Trial -> Packing Files

        Example:
        
        800 -> 0.44 -> 1Unjam -> configuration.asc
                                 radii.dat
        """
        subdirs = os.listdir(directory)
        paramData = {}
        for param in subdirs:
            paramDir = os.path.join(directory, param)
            if not os.path.isdir(paramDir): # Only look in dirs
                continue
            paramDF = tr.Trial(paramDir, nbs=nbs)
            if len(paramDF) <= 1: # Throw out dirs with no good data
                continue
            if bidispersity:
                label = (1 - float(param))/(1 + float(param))
            else:
                label = param
            paramData[label] = paramDF
            paramData = OrderedDict(sorted(paramData.items(), key=lambda t: t[0]))

        return paramData

    def calculatePF_all(self, report='mean', cutoff=0.0):
        """
        Calculates the packing fraction for the full dataset of the packings. 
        
        There are a number of different options that can be used for reporting
        the packing fractions. The mean of the trials is the default, but the
        median, max, min, and mean given a certain cutoff are all possible,
        where you supply what you want reported as an argument.

        Example:
        
        data.calculatePF_all('max')
        
        Will find the maximum order parameter for each bidispersity parameter.
        """
        pfs = {}
        for key, val in self.trials.items():
            val.calculatePF_all()
            if report == 'mean': 
                pfs[key] = val.df.pf.mean()
            elif report == 'median':
                pfs[key] = val.df.pf.median()
            elif report == 'max':
                pfs[key] = val.df.pf.max()
            elif report == 'min':
                pfs[key] = val.df.pf.min()
            elif report == 'cutoff':
                pfs[key] = val.df.pf[val.df.pf > cutoff].mean()
            else:
                print(report, ' not a valid statistic to report, reporting',
                      ' mean')
                pfs[key] = val.df.pf.mean()
        pfS = pd.Series(pfs, name='pf')
        self.df['pf'] = pfS

    def calculateOP_all(self, n=6, report='mean', cutoff=0.0):
        """
        Calculates the n-atic order parameter for the full dataset of packings,
        requires that the neighbor information for each packing be in the
        directory of each packing.

        There are a number of different options that can be used for reporting
        the order parameters. The mean of the trials is the default, but the
        median, max, min, and mean given a certain cutoff are all possible,
        where you supply what you want reported as an argument.

        Example:
        
        data.calculateOP_all('max')
        
        Will find the maximum order parameter for each bidispersity parameter.
        """
        ops = {}
        for key, val in self.trials.items():
            val.calculateOP_all(n)
            if report == 'mean': 
                ops[key] = val.df[str(n) + '-atic'].mean()
            elif report == 'median':
                ops[key] = val.df[str(n) + '-atic'].median()
            elif report == 'max':
                ops[key] = val.df[str(n) + '-atic'].max()
            elif report == 'min':
                ops[key] = val.df[str(n) + '-atic'].min()
            elif report == 'cutoff':
                ops[key] = val.df[str(n) + '-atic'][val.df[str(n) + '-atic'] >
                           cutoff].mean() 
            else:
                print(report, ' not a valid statistic to report, reporting',
                      ' mean')
                ops[key] = val.df[str(n) + '-atic'].mean()
        opS = pd.Series(ops, name=str(n) + '-atic')
        self.df[str(n) + '-atic'] = opS
    
    def calculateSegOP_all(self, n=6, report='mean', cutoff=0.0):
        """
        Calculates the segregated n-atic order parameter for the full dataset of
        packings, requires that the neighbor information for each packing be in
        the directory of each packing.

        There are a number of different options that can be used for reporting
        the order parameters. The mean of the trials is the default, but the
        median, max, min, and mean given a certain cutoff are all possible,
        where you supply what you want reported as an argument.

        Example:
        
        data.calculateSegOP_all('max')
        
        Will find the maximum segregated order parameter for each bidispersity
        parameter.
        """
        opsBig = {}
        opsLittle = {}
        for key, val in self.trials.items():
            val.calculateSegOP_all(n)
            if report == 'mean': 
                opsBig[key] = val.df[str(n) + '-aticBig'].mean()
                opsLittle[key] = val.df[str(n) + '-aticLittle'].mean()
            elif report == 'median':
                opsBig[key] = val.df[str(n) + '-aticBig'].median()
                opsLittle[key] = val.df[str(n) + '-aticLittle'].median()
            elif report == 'max':
                opsBig[key] = val.df[str(n) + '-aticBig'].max()
                opsLittle[key] = val.df[str(n) + '-aticLittle'].max()
            elif report == 'min':
                opsBig[key] = val.df[str(n) + '-aticBig'].min()
                opsLittle[key] = val.df[str(n) + '-aticLittle'].min()
            elif report == 'cutoff':
                opsBig[key] = val.df[str(n) + '-aticBig'][val.df[(str(n) +
                              '-aticBig')] > cutoff].mean() 
                opsLittle[key] = val.df[str(n) + '-aticLittle'][val.df[(str(n) +
                              '-aticLittle')] > cutoff].mean() 
            else:
                print(report, ' not a valid statistic to report, reporting',
                      ' mean')
                opsBig[key] = val.df[str(n) + '-aticBig'].mean()
                opsLittle[key] = val.df[str(n) + '-aticLittle'].mean()
        opSBig = pd.Series(opsBig, name=str(n) + '-aticBig')
        opSLittle = pd.Series(opsLittle, name=str(n) + '-aticLittle')
        self.df[str(n) + '-aticBig'] = opSBig
        self.df[str(n) + '-aticLittle'] = opSLittle

    def calculateDFG_all(self, report='mean'):
        """
        Calculates the size of the defect subgraph for the entire dataset.

        There are a number of different options that can be used for reporting
        the order parameters. The mean of the trials is the default, but the
        median, max, and min. Where you supply what you want reported as an
        argument.

        Example:
        
        data.calculateDFG_all('max')
        
        Will find the maximum DFG size for each bidispersity parameter.
        """
        dfgs = {}
        for key, val in self.trials.items():
            val.calculateDFG_all()
            if report == 'mean': 
                dfgs[key] = val.df['DFGS'].mean()
            elif report == 'median':
                dfgs[key] = val.df['DFGS'].median()
            elif report == 'max':
                dfgs[key] = val.df['DFGS'].max()
            elif report == 'min':
                dfgs[key] = val.df['DFGS'].min()
            else:
                print(report, ' not a valid statistic to report, reporting',
                      ' mean')
                dfgs[key] = val.df['DFGS'].mean()
        dfgS = pd.Series(dfgs, name='DFGS')
        self.df['DFGS'] = dfgS

    def calculateDFF_all(self, report='mean', cutoff=1.0):
        """
        Calculates the defect fraction for the entire dataset.

        There are a number of different options that can be used for reporting
        the order parameters. The mean of the trials is the default, but the
        median, max, and min. Where you supply what you want reported as an
        argument.

        Example:
        
        data.calculateDFF_all('max')
        
        Will find the maximum DFF size for each bidispersity parameter.
        """
        dffs = {}
        for key, val in self.trials.items():
            val.calculateDFF_all()
            if report == 'mean': 
                dffs[key] = val.df['DFF'].mean()
            elif report == 'median':
                dffs[key] = val.df['DFF'].median()
            elif report == 'max':
                dffs[key] = val.df['DFF'].max()
            elif report == 'min':
                dffs[key] = val.df['DFF'].min()
            elif report == 'cutoff':
                dffs[key] = val.df['DFF'][val.df['DFF'] < cutoff].mean()
            else:
                print(report, ' not a valid statistic to report, reporting',
                      ' mean')
                dffs[key] = val.df['DFF'].mean()
        dffS = pd.Series(dffs, name='DFF')
        self.df['DFF'] = dffS

    def calculateCM_all(self, report='mean', cutoff=0.0):
        """
        Calculates the coordination matrix for the entire dataset.

        There are a number of different options that can be used for reporting
        the order parameters. The mean of the trials is the default, but the
        median, max, and min. Where you supply what you want reported as an
        argument.

        Example:
        
        data.calculateCM_all('max')
        
        Will find the maximum CM value for each bidispersity parameter.

        """
        cmsbb = {}; cmsbl = {}; cmslb = {}; cmsll = {}
        for key, val in self.trials.items():
            val.calculateCM_all()
            if report == 'mean': 
                cmsbb[key] = val.df['BB'].mean()
                cmsbl[key] = val.df['BL'].mean()
                cmslb[key] = val.df['LB'].mean()
                cmsll[key] = val.df['LL'].mean()
            elif report == 'median':
                cmsbb[key] = val.df['BB'].median()
                cmsbl[key] = val.df['BL'].median()
                cmslb[key] = val.df['LB'].median()
                cmsll[key] = val.df['LL'].median()
            elif report == 'max':
                cmsbb[key] = val.df['BB'].max()
                cmsbl[key] = val.df['BL'].max()
                cmslb[key] = val.df['LB'].max()
                cmsll[key] = val.df['LL'].max()
            elif report == 'min':
                cmsbb[key] = val.df['BB'].min()
                cmsbl[key] = val.df['BL'].min()
                cmslb[key] = val.df['LB'].min()
                cmsll[key] = val.df['LL'].min()
            elif report == 'cutoff':
                cmsbb[key] = val.df['BB'][val.df['BB'] > cutoff].mean()
                cmsbl[key] = val.df['BL'][val.df['BL'] > cutoff].mean()
                cmslb[key] = val.df['LB'][val.df['LB'] > cutoff].mean()
                cmsll[key] = val.df['LL'][val.df['LL'] > cutoff].mean()
            else:
                print(report, ' not a valid statistic to report, reporting',
                      ' mean')
                cmsbb[key] = val.df['BB'].mean()
                cmsbl[key] = val.df['BL'].mean()
                cmslb[key] = val.df['LB'].mean()
                cmsll[key] = val.df['LL'].mean()
        self.df['BB'] = pd.Series(cmsbb)
        self.df['BL'] = pd.Series(cmsbl)
        self.df['LB'] = pd.Series(cmslb)
        self.df['LL'] = pd.Series(cmsll)
