#!/usr/bin/python

import re
import numpy as np
import math
from optparse import OptionParser
import sys
import os

eps = 1.0e-20

def importNeighborList(importFile):
    """
    This function takes in the tess_tri.dat file created by the radical
    tesselation package and pulls out the list of neighbors from it. 

    Pass in the file object containing the tess_tri.dat output and it will
    return a list of length number of particles that contains each particle's
    neighbor. BE WARY OF THE FACT THAT THE PARTICLES IN THE WOLFRAM LANGUAGE
    OUTPUT ARE INDEXED FROM 1 NOT 0. In order to properly calculate the order
    parameters, this must be accounted for by subtracting 1 from the index
    whenever accessing the configuration or radii files.  

    Example for particle 1:

    Input:
    {{1, 123}, {1, 523}, {23, 1}, {1, 99}}

    Output:
    [123, 523, 23, 99]
    """
    data = importFile.read().replace('\n', '')

    # The tess_tri.dat file has is a list of two lists, the second one is the
    # triangulation that we want, so we use rfind to get it
    triStartIndex = data.rfind('{{{')

    tri = data[triStartIndex:]
    tri = tri.split('}},') # One line per particle

    newTri = []
    for i, part in enumerate(tri): #Remove junk chars and self referencing
        newLine = re.findall(r'\d+', part)
        newLine = [int(i) for i in newLine]
        newLine = list(filter(lambda x: x != i + 1, newLine))
        
        newTri.append(newLine)
    
    return newTri

def cross(a, b):
    """
    Returns a cross b.
    """

    return np.array([a[1]*b[2] - a[2]*b[1], a[2]*b[0] - a[0]*b[2], a[0]*b[1] -
                    a[1]*b[0]])

def neighborCoordinates(particleNum, coordList, neighborList):
    """
    Given a certain particle and the list of coordinates and neighbors, this
    function generates the list of a particle's neighbors' coordinates. The fact
    that the indexing of the neighbor list is off, is taken into account in this
    function.

    Inputs: particleNum - index of desired particle
            coordList - list of all particle coordinates in the packing
            neighborList - list of all particle neighborhoods in the packing

    Returns: numpy array of the neighbor's coordinates
    """
    return list(map(lambda x : coordList[x], neighborList))

def tangentPlaneProjection(centralPoint, neighborPoint):
    """
    Given a point on a sphere and neighboring point, computes the projection of
    the neighboring point onto the central point's tangent plane.
    
    Inputs: numpy arrays of coordinates

    Return: Projected point
    """
    dot = np.dot(centralPoint, neighborPoint)
    scale = np.dot(centralPoint, centralPoint) / dot
    return scale*neighborPoint


def coplanarDimensionReduction(centralPoint, projectedNeighbors):
    """
    Given a central particle and the projection of its neighbors onto the
    tangent plane. Reduces the dimensionality of the coordinates by getting 2D
    coordinates of the neighbors withing the tangent plane

    Inputs: centralPoint - Central particle coordinate
            projectedNeighbors - List with projected neighbor coords

    Returns: List of neighbor coordinates reduced to two dimensions
    """
    normVec = centralPoint / np.dot(centralPoint, centralPoint)**(0.5)
    if len(projectedNeighbors) < 1: # No neighbors
        return projectedNeighbors
    xHat = projectedNeighbors[0] - centralPoint
    yHat = cross(xHat, normVec)

    transformedCoords = []
    for part in projectedNeighbors:
        newCoords = np.array([np.dot(part, xHat), np.dot(part, yHat)])
        transformedCoords.append(newCoords)

    return transformedCoords

def getNeighborsAndProject(partNum, coordList, neighborList):
    """
    Function that essentially combines the various projections and neighbor
    grabbing functions into one. This allows the user to simply pass the
    particle number for which they are trying to get the correct projection and
    it will return the two dimensional coordinates of the neighbors in the
    projection.

    Inputs: partNum - desired particle
            coordList - Numpy array of coordinates in 3D
            neighborList - list of particle neighbors

    Returns: List of coordinates of the projected neighbors onto the two
    dimensional tangent plane
    """
    neighborCoords = neighborCoordinates(partNum, coordList, neighborList)
    projectedNeighbors = []
    for part in neighborCoords:
        projectedNeighbors.append(tangentPlaneProjection(coordList[partNum], 
                                  part))

    projectedNeighbors = coplanarDimensionReduction(coordList[partNum],
                         projectedNeighbors)

    return projectedNeighbors

def binomial(n, k):
    """
    Returns n choose k
    """
    return math.factorial(n) / (math.factorial(n - k)*math.factorial(k))

def calculateSingleOrderParameter(n, x, y, bList):
    """
    Given the x and y coordinates of a neighboring particle on the tangent plane
    (the central particle is assumed to be at the origin), and which of the
    n-atic order paramters are being calculated (input by n), this computes the
    single particle contribution to the order parameter.

    Inputs: n - integer determining which order parameter it is
            x, y - floating point numbers for coords

    Returns: numpy array containing the real and imaginary parts
    """
    realPart = 0.
    imPart   = 0.
    normNegN = (x**2 + y**2)**(-n*0.5)

    for i in range(n + 1):
        val = bList[i]*normNegN*(x**(n - i))*(y**i)
        if not i % 4: # Divisible by 4
            realPart += val
        elif not i & 1: # Even (only called when not divisible by 4)
            realPart += -val
        elif not (i + 3) % 4: 
            imPart += val
        else:
            imPart += -val

    return np.array([realPart, imPart])

def calculateOrderParameter(n, neighbors):
    """
    Given the order parameter that you want to calculate, n. (Ex: hexatic -> n =
    6) and a list of neighbors, calculates the order parameter.

    Inputs: List of neighbors, whose coordinates are transformed to two
    dimensions on the tangent plane and passed as numpy arrays [[x1, y1], ...,
    [xN, yN]]. 
    n in the n-atic order parameter

    Returns: Order parameter value
    """
    bList = [binomial(n, k) for k in range(n + 1)]
    orderParam = np.array([0., 0.])
    for part in neighbors:
        orderParam += calculateSingleOrderParameter(n, part[0], part[1], bList)
    return np.dot(orderParam, orderParam)**(0.5) / len(neighbors)
