import numpy as np
import pandas as pd
import os
import patools.op as op
import re

class Packing:
    __configFileName   = 'configuration.asc'
    __radiiFileName    = 'radii.dat'
    __progressFileName = 'progress.txt'
    __neighborFileName = 'tess_tri.dat'
    
    __jammedIndicator  = 'consecutive jammed steps: 10'
    
    def __init__(self, directory=os.getcwd(), nbs=False):
        self.df = self.loadPacking(directory)
        if nbs:
            self.importNeighbors(directory)
        self.jammed    = self._finishedJamming(os.path.join(directory,
                         self.__progressFileName))
        self.pf = 0.0
        self.op = {}
        self.ops = {}

    def __len__(self):
        return len(self.df)
    
    def _loadConfig(self, configFileLoc):
        try:
            with open(configFileLoc, 'r') as cf:
                # Below tests if letters are in the first 9 chars
                # Will be true if Mathematica output was broken
                # Idicates a garbage config file
                if not re.search('[A-Za-z]', cf.read(9)): 
                    cf.seek(0) # Reset file pointer
                    configDF = pd.read_table(cf, header=None,
                            delim_whitespace=True, names=['x','y','z'])
                else: 
                    configDF = pd.DataFrame(columns=['x','y','z'])
        except OSError:
            print("Config file not found")
            configDF = pd.DataFrame()

        return configDF

    def _loadRadii(self, radiiFileLoc):
        try:
            radiiDF = pd.read_table(radiiFileLoc, header=None, names=['radii'])
        except OSError:
            print("Radii file not found")
            return pd.Series()

        return radiiDF
    
    def _finishedJamming(self, progressFileLoc):
        try:
            with open(progressFileLoc, 'r') as pf:
                for line in pf:
                    pass
                last = line
                return (self.__jammedIndicator in last)
        except:
            return False # Assume packings are not jammed by default
    
    def _packingToDataframe(self, coords, radii):
        df = coords
        df['radii'] = radii

        return df
    
    def loadPacking(self, directory=os.getcwd()):
        """
        Imports from the given directory the various files defining a packing into a
        pandas dataframe. The default directory to look in is the current working
        directory, but a path can be specified.
        """
        coords = self._loadConfig(os.path.join(directory, self.__configFileName))
        radii  = self._loadRadii (os.path.join(directory, self.__radiiFileName ))
    
        return self._packingToDataframe(coords, radii)

    def setConfigName(self, newName):
        """
        Modifies the default configuration file name from configuration.asc to
        whatever is provided.
        """
        self.__configFileName = newName
        
    def setRadiiName(self, newName):
        """
        Modifies the default radii file name from radii.dat to whatever is provided.
        """
        self.__radiiFileName = newName

    def calculatePF(self):
        """
        Computes the packing fraction of a given packing
        """
        if self.pf == 0.0:
            singlePart = self.df.loc[0]

            sphereRad = ((singlePart['x']**2 + singlePart['y']**2 +
                         singlePart['z']**2)**(0.5))

            coveredArea = 0.0
            for p in self.df.radii.values:
                coveredArea += p**2

            self.pf = coveredArea / (4*sphereRad**2)

        return self.pf 

    def importNeighbors(self, directory=os.getcwd()):
        """
        From the tess_tri.dat file and using the functions defined in the
        repurposed order params code, imports the neighbor list and puts
        neighbor information into the packing object. 
        """
        try:
            with open(os.path.join(directory, self.__neighborFileName), 
                                                                     'r') as nb:
                importedNB = op.importNeighborList(nb)
                self.df['coordination number'] = [len(i) for i in importedNB]
                adjNB = list(map(lambda x: ' '.join([str(i - 1) for i in x]),
                        importedNB))
                self.df['nList'] = adjNB
        except:
            self.df['coordination number'] = ''
            self.df['nList'] = ''

    def calculateOP(self, n=6):
        """
        Using the order parameter code in op.py, calculates the n-atic order
        parameter for each particle and stores it in the dataframe. Also
        computes the average and stores it as part of the object in a dict. The
        dict allows for the n-atic order parameter to be called upon, so
        self.op[6] would be the mean hexatic order parameter for the packing.
        """
        coords = self.df[['x', 'y', 'z']].as_matrix()
        orderParams = []
        nListFull = self.df.nList.values
        for i in range(len(coords)):
            nbList = [int(x) for x in nListFull[i].split()]
            nbCoords = op.getNeighborsAndProject(i, coords, nbList)
            orderParams.append(op.calculateOrderParameter(n, nbCoords))
        self.df[str(n) + '-atic'] = orderParams
        self.op[n] = sum(orderParams) / len(orderParams)

    def calculateSegOP(self, n=6):
        """
        Using the order parameter code in op.py, calculates the n-atic order
        parameter for each particle and all of its neighbors of the same size
        and stores it in the dataframe. Also computes the average and stores it
        as part of the object in a dict. The dict allows for the n-atic order
        parameter to be called upon, so self.op[6] would be the mean hexatic
        order parameter for the packing.
        """
        coords = self.df[['x', 'y', 'z']].as_matrix()
        numBig = (self.df.radii != self.df.radii[0]).idxmax() 
        orderParams = []
        nListFull = self.df.nList.values
        for i in range(len(coords)):
            if i < numBig:
                nbList = [int(x) for x in nListFull[i].split() 
                          if int(x) < numBig]
            else:
                nbList = [int(x) for x in nListFull[i].split() 
                          if int(x) >= numBig]
            nbCoords = op.getNeighborsAndProject(i, coords, nbList)
            orderParams.append(op.calculateOrderParameter(n, nbCoords))
        self.df[str(n) + '-aticSeg'] = orderParams
        bigOP = self.df[str(n) + '-aticSeg'].loc[:numBig-1]
        littleOP = self.df[str(n) + '-aticSeg'].loc[numBig:]
        self.ops[str(n) + 'B'] = bigOP[bigOP.notnull()].mean()
        self.ops[str(n) + 'L'] = littleOP[littleOP.notnull()].mean() 

    def calculateDefectGraphSize(self):
        """
        Calculates the size of the defect subgraph, namely N - N_6.
        """
        dfG = sum(self.df['coordination number'] != 6)
        self.dfG = dfG

    def calculateDefectFraction(self):
        """
        Calculates 1 - N_6 / N for a given packing.
        """
        nb6 = sum(self.df['coordination number'] == 6)
        self.dfN =  1 - nb6 / len(self.df)

    def calculateCoordinationMatrix(self):
        """
        Calculates the coordination matrix analysis for a packing. This records
        how many neighbor pairs are big-big, big-little, and little-little.
        """
        self.cm = np.array([0., 0., 0., 0.])

        numBig = (self.df.radii != self.df.radii[0]).idxmax() 
        numSmall = len(self.df) - numBig

        nListFull = self.df.nList.values

        for i in range(len(self.df)):
            nList = nListFull[i].split()
            nList = [(int(x) < numBig) for x in nList]

            big = sum(nList)
            small = len(nList) - big
            if i < numBig:
                self.cm[0] += big
                self.cm[1] += small
            else:
                self.cm[2] += big
                self.cm[3] += small

        self.cm[0] = self.cm[0] / numBig
        self.cm[1] = self.cm[1] / numBig
        self.cm[2] = self.cm[2] / numSmall
        self.cm[3] = self.cm[3] / numSmall
