import os
import pandas as pd
import patools.packing as pck

class Trial:

    def __init__(self, directory=os.getcwd(), nbs=False):
        self.packings = self.loadTrials(directory, nbs)
        self.df = pd.DataFrame(index=self.packings.keys())

    def __len__(self):
        return sum([len(packing) for packing in self.packings.values()])

    def loadTrials(self, directory, nbs=False):
        """
        Loads a set of packings corresponding to one parameter set. The
        subdirectories should contain packings files themselves.

        ParamDirectory -> Trial Subdirs -> Packing Files
        """
        subdirs = os.listdir(directory)
        trialPackings = {}
        for trial in subdirs:
            trialDir = os.path.join(directory, trial)
            if not os.path.isdir(trialDir): # Only look in dirs
                continue

            newPacking = pck.Packing(trialDir,nbs=nbs)
            if len(newPacking) <= 1: # Remove broken packings
                continue
            trialPackings[trial] = newPacking

        return trialPackings

    def calculatePF_all(self):
        """
        Calculates the packing fraction for each packing in a trial. The packing
        fractions for each trial are stored in a dataframe. This function is
        most often accessed from the dataset object, where different statistics
        of the packing fractions are reported, such as the mean, median, max, or
        min.
        """
        pfs = {}
        for key, val in self.packings.items():
            pfs[key] = val.calculatePF()
        pfS = pd.Series(pfs, name='pf')
        self.df['pf'] = pfS
        return pfs

    def calculateOP_all(self, n=6):
        """
        Calculates the n-atic order parameter for each packing in a trial. The
        average order parameter for a packing is stored in a dataframe. This
        function is most often accessed from the dataset object, where different
        statistics of the packing fractions are reported, such as the mean,
        median, max, or min.
        """
        ops = {}
        for key, val in self.packings.items():
            val.calculateOP(n)
            ops[key] = val.op[n]
        opS = pd.Series(ops, name=str(n) + '-atic')
        self.df[str(n) + '-atic'] = opS
        return opS

    def calculateSegOP_all(self, n=6):
        """
        Calculates the segregated n-atic order parameter for each packing in a
        trial. The average order parameter for big-big and little-little for a
        packing is stored in a dataframe. This function is most often accessed
        from the dataset object, where different statistics of the packing
        fractions are reported, such as the mean, median, max, or min.
        """
        opsBig = {}
        opsLittle = {}
        for key, val in self.packings.items():
            val.calculateSegOP(n)
            opsBig[key] = val.ops[str(n) + 'B']
            opsLittle[key] = val.ops[str(n) + 'L']
        opSBig = pd.Series(opsBig, name=str(n) + '-aticBig')
        opSLittle = pd.Series(opsLittle, name=str(n) + '-aticLittle')
        self.df[str(n) + '-aticBig'] = opSBig
        self.df[str(n) + '-aticLittle'] = opSLittle

    def calculateDFG_all(self):
        """
        Calculuates the defect graph size for all of the packings in a given
        trial.
        """
        dfgs = {}
        for key, val in self.packings.items():
            val.calculateDefectGraphSize()
            dfgs[key] = val.dfG
        dfgS = pd.Series(dfgs, name='DFGS')
        self.df['DFGS'] = dfgS
        return dfgS

    def calculateDFF_all(self):
        """
        Calculates the defect fraction for all of the packings in a given trial.
        """
        dffs = {}
        for key, val in self.packings.items():
            val.calculateDefectFraction()
            dffs[key] = val.dfN
        dffS = pd.Series(dffs, name='DFF')
        self.df['DFF'] = dffS
        return dffS

    def calculateCM_all(self):
        """
        Calculates the coordination matrix for all of the packings in a given
        trial.
        """
        cmsbb = {}; cmsbl = {}; cmslb = {}; cmsll = {}
        for key, val in self.packings.items():
            val.calculateCoordinationMatrix()
            cmsbb[key] = val.cm[0]
            cmsbl[key] = val.cm[1]
            cmslb[key] = val.cm[2]
            cmsll[key] = val.cm[3]
        self.df['BB'] = pd.Series(cmsbb)
        self.df['BL'] = pd.Series(cmsbl)
        self.df['LB'] = pd.Series(cmslb)
        self.df['LL'] = pd.Series(cmsll)


